/***************************************************************************
 *   Copyright (C) 2005 by Bram Geron                                      *
 *   bgeron@planet.nl                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>
#include <string>
#include <utility>
#include <vector>
#include <cstring>
#include <cstdio>
#include <sstream>
#include <fstream>

using namespace std;
// How long the length of a terminal is supposed to be
const int TEXT_WIDTH = 60;
const int BUF_SIZE = 1024;

int bytecounter=0;

typedef vector<pair<string,string> > entry;
typedef vector<entry> definitions;
typedef pair<string,int> indexentry;
typedef vector<indexentry> Index;

stringstream indexfile_buffer,
             deffile_buffer;

class wrong_line : exception {};

char chars_used[256] = { // Which characters are used
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,
false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false
};
             
char* base64_encode(unsigned int number) {

  // Returns the base64 of number
  
  if (!number) return "A";
  
  const char* letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  
  int storage=2;
  unsigned int temp1 = 0xffffffc0, temp2 = 0x3F;
  // If there are bits in common with temp1, we need another byte.
  // temp2 represents 1 byte storage
  while (number & temp1) { storage++; temp2 <<= 6; temp1 -= temp2; }
  char* tempstr = new char[storage];
  
  tempstr[--storage]='\0';
  
  while (storage--) {
    tempstr[storage] = letters[number & 0x3f];
    
    number >>=6;
  }
  return tempstr;
  
}

void storeentry(const string &index, const string &value, const char* const to_be_appended="\n\n") {
  // Store something
  
  // Add index to the characters allowed
  for (string::const_iterator i = index.begin();i<index.end();i++)
    chars_used[*i] = true;
  
  // Save the starting position
  int begin = bytecounter;
  
  // Store the actual data
  deffile_buffer << value.c_str() << to_be_appended << flush;
  
  bytecounter += value.size()+strlen(to_be_appended);
  
  // Store the index in the indexbuffer
  indexfile_buffer << index << '\t' << 
    base64_encode(begin) << '\t' << 
    base64_encode(bytecounter-begin) << '\n' << flush;
  
}

string format(const string& input_str,int width, int indent) {
  
  // Format text not to exceed width columns (word-wrap)
  
  stringstream input(input_str),output,buffer;
  string buffer_str, whitespace(" \t"), output_str;
  int pos=0;
  bool firstflush=true;
  
  while(input.good()) {
    char character = input.get();
    if (character == EOF) continue;
    if (whitespace.find(character)!=string::npos) {
      if (pos+buffer.str().size()+1 > width) {
        output << '\n' << string(indent,' '); pos=0;
      } else if (!firstflush) output << ' ';
      else firstflush=false;
      
      output << buffer.str();
      buffer.str(string("")); // empty the buffer
    } else {
      buffer << character;
      pos++;
    }
  }

  // Flush the remaining buffer.
  if (pos+buffer.str().size()+1 > width) {
    output << '\n' << string(indent,' '); pos=0;
  } else if (!firstflush) output << ' ';
  
  output << buffer.str();
  buffer.str(string("")); // empty the buffer
  
  return output.str();

}

void storeentry(const string &index, const entry &content) {
  // Store an entry
  
  string whattostore;
  
  // Determine the longest name of a property of an entry
  int indentation=0;
  for (entry::const_iterator begin=content.begin();begin<content.end();begin++) {
    if (begin->first.size()>indentation) indentation=begin->first.size();
  }
  
  // Store the properties
  for (entry::const_iterator begin=content.begin();begin<content.end();begin++)
    whattostore.append(string(indentation-begin->first.size(),' ') + begin->first + ": " + format(begin->second,TEXT_WIDTH-indentation-2, indentation+2) + '\n') ;
  
  storeentry(index,whattostore);
}

void read(istream &stream, char* dest, int maxchars=0, char delim='\n') {
  if (maxchars) {
    char currentchar;
    while (stream.good() and --maxchars and (currentchar = stream.get())!=delim)
      *dest++ = currentchar;
  } else {
    char currentchar;
    while (stream.good() and (currentchar = stream.get())!=delim)
      *dest++ = currentchar;
  }
}

bool israfsi (const string &str) {
  string::const_iterator begin = str.begin();
  while (begin < str.end()) {
    if (*begin>='A' and *begin <='Z') return false;
    if (*begin>='a' and *begin <='z' and *begin != 'h') return true;
    begin++;
  }
  throw wrong_line();
}

int main(int argc, char *argv[]) {

  if (argc<2) {cerr << 
argv[0] << " makes a dict file out of a lojban tex file.\n"
"Syntax: " << argv[0] <<  " texfile [url [short name]] \n\n"; exit (1);}

  // Open the necessary files
  fstream texfile(argv[1]);
  
  char* pos=argv[1];
  while (*pos!='\0' and *pos != '.') pos++;
  string base_filename(argv[1],pos);
  time_t *currenttime = new time_t(time(0));
  storeentry("00-database-info",string(
"This database was converted from tex to TEI format\n"
"using lojtexdict.\n"
"Converter copyright (c) 2005, Bram Geron.\n"
"The database is public domain.\n"
"Date of convertion: ")+ctime(currenttime)+"\n\n");
  delete currenttime;

  if (argc>2) storeentry("00-database-short",argv[2]);
  if (argc>3) storeentry("00-database-url",argv[3]);
  
  try {
  while (texfile.good()) {
/*  for (texfile.getline(linebufferstr,BUF_SIZE),linebuffer.str(linebufferstr);
       texfile.good();
       texfile.getline(linebufferstr,BUF_SIZE),linebuffer.str(linebufferstr)) {*/
    stringstream linebuffer,wordbuffer;
    char linebufferstr[BUF_SIZE];
  
  
    texfile.getline(linebufferstr,BUF_SIZE),linebuffer.str(linebufferstr);
       
    // Check if this line is a valid definition
    char* check = new char[8];
    check[8]='\0';
    
    linebuffer.readsome(check,8);
    if (!linebuffer.good() or string(check)!="\\textbf{") {delete check;continue;}
    
    entry currententry;
    
    // Read the name of the word
    string name;
    getline(linebuffer,name,'}');
    currententry.push_back(make_pair(string("Name"),name));
    
    if (name=="a'a") {
    new char;
    }
        
    // We got a problem if there's no "} \markboth{<name>}{<name>}" in the line
    check = new char[16+2*name.size()];
    linebuffer.get(check,15+2*name.size());
    if (check!=" \\markboth{"+name+"}{"+name+"}") {
      delete check;
      throw wrong_line();
    }
    delete check;
    
    bool defalreadystarted=false;
    
    while (linebuffer.good()) {
      switch (char character = linebuffer.get()) {
      case EOF:
        break;
      case '$':
        // Something like $x_1$ starts
        while ((character = linebuffer.get())!='$' and linebuffer.good())
          switch (character) {
          case '_':case '{':case '}':break;
          default: wordbuffer << character;
          }
        defalreadystarted=true;
        break;
      case '\\':
      
        switch (char peek = linebuffer.peek()) {// which escapes are allowed
        case '&':
        case '%':
        case '#':
          break;
        case 't':
          {
          
            // There's either a rafsi or a selma'o specified
            check = new char[6];
            check[5]='\0';
            linebuffer.get(check,5);
            linebuffer.get();linebuffer.get();
            if (strcmp(check,"text") or linebuffer.get()!='{') // Problemo
              {delete check;throw wrong_line();}
            delete check;
            
            character = linebuffer.get();
            if (character == '[') {
              string raforsel;
              getline(linebuffer,raforsel,']');
              linebuffer.get();
              
              // Determine if it's a rafsi or a selma'o
              if (israfsi(raforsel))
                currententry.push_back(make_pair(string("rafsi"),raforsel));
              else currententry.push_back(make_pair(string("selma'o"),raforsel));
              continue;
            } else if (character == '(') {
              string inthemeaning;
              getline(linebuffer,inthemeaning,')');
              linebuffer.get();
              currententry.push_back(make_pair(string("In the meaning"),inthemeaning));
              continue;
            }
          }
        default:
          throw wrong_line();// Escape not allowed
        }
      
      case ' ':
        if (!defalreadystarted) break; // All entries start with a space, remove it
      default:
        defalreadystarted=true;
        wordbuffer << character;
      }
    }
    
    // Store the definition
    currententry.push_back(make_pair(string("Definition"),wordbuffer.str()));
    
    // Store the entry
    storeentry(name,currententry);
  } // end for
  } catch (wrong_line) {
    cerr << "Invalid line at bytepos " << texfile.tellg(); exit (1);
  }
  
  // Add an entry for the characters allowed
  string charsallowed;
  for (int i=0;i<256;i++)
    if (chars_used[i]) charsallowed += char(i);
  storeentry("00-database-alphabet",charsallowed,"\n");
  storeentry("00-database-allchars","00-database-allchars");
  
  ofstream deffile(string(base_filename+".dict").c_str());
  deffile << deffile_buffer.rdbuf();
  deffile.close();
  system(string("dictzip "+base_filename+".dict").c_str());
  
  ofstream indexfile("indexfile_temp");
  indexfile << indexfile_buffer.rdbuf();
  indexfile.close();
  system(string("sort -df < indexfile_temp > " + base_filename + ".index").c_str());
  
  exit(EXIT_SUCCESS);
}
